/*****************************************************************
**
**	cal.c
**
**	print calendar entries as defined by RFC5545 and RFC7986
**
*****************************************************************/

# include <stdio.h>
# include <time.h>
# include <assert.h>
# include <string.h>
# include <ctype.h>

# include "config.h"
# include "vcl.h"
# include "vocab.h"
#define extern
# include "cal.h"
#undef extern

# define	DAYSEC(n)	((n) * 24 * 3600)

static	char	*datestr (struct tm *date, int withtime);
static	char	*isodatestr (struct tm *date, int withtime);
static	time_t	firstday;

/*****************************************************************
**	setstartday (datestr)
**	
**	This function is to store the starting date into the local
**	static var 'firstday'.
**	If called w/o a starting date string (dd.mm.yyyy) the current
**	date will be used to calculate the last monday from.
**	
**	returns seconds of the epoch
*****************************************************************/
time_t	setstartday (const char *datestr)
{
	struct	tm	date;
	time_t		sec;

	sec = 0L;
	if ( datestr && *datestr )
		sec = startdate (datestr);
# if 0
	firstday = day (&date, sec);
#else
	// firstday = nextmonday (&date, sec);
	firstday = lastmonday (&date, sec);
#endif

	return firstday;
}

/*****************************************************************
**	day (date, time_t sec)
**	Fills out the tm struct with the values appropriate to the
**	current or given time. 
**	We use UTC here to overcome problems with summertime changes.
**	returns seconds of the epoch
*****************************************************************/
time_t	day (struct tm *date, time_t sec)
{
	assert ( date != NULL );

	if ( sec == 0 )
		time (&sec);
	
#if 1
	gmtime_r (&sec, date);
#else
	localtime_r (&sec, date);
	date->tm_hour = 8;
	date->tm_min = date->tm_sec = 0;
#endif

	return sec;
}

/*****************************************************************
**	nextmonday (date, time_t sec)
**	returns seconds of the epoch
*****************************************************************/
time_t	nextmonday (struct tm *date, time_t sec)
{
	int		wday;

	assert ( date != NULL );

	sec = day (date, sec);
	wday = date->tm_wday;		/* sun == 0 */

	return day (date, sec + DAYSEC( ((7 - wday)%7) + 1 ));
}

/*****************************************************************
**	lastmonday (date, time_t sec)
**	returns seconds of the epoch
*****************************************************************/
time_t	lastmonday (struct tm *date, time_t sec)
{
	int		wday;

	assert ( date != NULL );

	sec = day (date, sec);
	wday = date->tm_wday;		/* sun == 0 */

	return day (date, sec - DAYSEC( ((5 + wday)%7) + 1 ));
}

/*****************************************************************
**
**	startdate (char *datestr)
**
**	Take a date string in the form "dd.mm.yyyy" and calculates
**	the number of seconds since the epoch.
**
**	returns seconds of the epoch
**
*****************************************************************/
time_t	startdate (const char *datestr)
{
	int		day;
	int		mon;
	int		year;
	struct	tm	date;
	time_t		sec;

	assert ( datestr != NULL );

	if ( sscanf (datestr, "%d.%d.%d", &day, &mon, &year) == 3 )
	{
		if ( day >= 1 && day <= 31 && mon >= 1 && mon <= 12 && year > 1900 )
		{
			date.tm_mday = day;
			date.tm_mon = mon - 1;
			date.tm_year = year - 1900;
			date.tm_hour = 8;
			date.tm_min = 0;
			date.tm_sec = 0;
			sec = mktime (&date);
		}
		else
			time (&sec);
	}
	else
		time (&sec);

	return sec;
}

/*****************************************************************
**	isodatestr ()
*****************************************************************/
static	char	*isodatestr (struct tm *date, int withtime)
{
	static	char	str[16+1];

	snprintf (str, sizeof (str), "%04d%02d%02dT%02d%02d%02dZ",
		date->tm_year + 1900, date->tm_mon + 1, date->tm_mday,
		date->tm_hour, date->tm_min, date->tm_sec);

	if ( !withtime )
		str[8] = '\0';

	return str;
}

/*****************************************************************
**	datestr ()
*****************************************************************/
static	char	*datestr (struct tm *date, int withtime)
{
	static	char	str[27+1];
	static	char	*wday[] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
	int	n;

	n = 0;
	if ( withtime )
		n += snprintf (str, sizeof (str), "%3s ", wday[date->tm_wday]);

	snprintf (str+n, sizeof (str) - n, "%02d.%02d.%04d %02d:%02d",
		date->tm_mday, date->tm_mon + 1, date->tm_year + 1900,
		date->tm_hour, date->tm_min);

	if ( !withtime )
		str[10] = '\0';

	return str;
}

/*****************************************************************
**	cal_start ()
*****************************************************************/
int	cal_start (FILE *out)
{
	int	n = 0;
	struct	tm	now;

	day (&now, 0L);

	n += fprintf (out, "BEGIN:VCALENDAR\r\n");
	n += fprintf (out, "VERSION:2.0\r\n");
	n += fprintf (out, "PRODID:-//HZNET//vocab//EN\r\n");
	n += fprintf (out, "METHOD:PUBLISH\r\n");
	n += fprintf (out, "NAME:%s Vocabulary\r\n", vocabname);
	//n += fprintf (out, "LAST-MODIFIED;VALUE=DATE-TIME:%s\r\n", isodatestr (&now, 1));
	n += fprintf (out, "LAST-MODIFIED:%s\r\n", isodatestr (&now, 1));
	n += fprintf (out, "REFRESH-INTERVAL;VALUE=DURATION:%s\r\n", CAL_REFRESH);
	if ( url )
		n += fprintf (out, "SOURCE;VALUE=URI:%s\r\n", url);

	return n;
}

/*****************************************************************
**	cal_end ()
*****************************************************************/
int	cal_end (FILE *out)
{
	int	n = 0;

	n += fprintf (out, "END:VCALENDAR\r\n");

	return n;
}

/*****************************************************************
**
**	print_folded ()
**
**	print a cal value of type text and fold lines which are
**	longer than 75 chars. This is a requirement from RFC5545.
**	Lines can be split at any position, but to avoid a split
**	in between an UTF8 character sequence, the split will be
**	made on white space chars.
**
**	'n' is the number of chars already printed on the line
**	'text' of course is the text to print.
**	A continous line will start with a tab character
**
*****************************************************************/
static	int	printfolded (FILE *out, int n, const char *text)
{
	int	i;
	int	len;
	int	maxc;

	maxc = 75 - n;		/* # of chars left on the current line */
	len = strlen (text);	/* length of the text */
	while ( len > maxc )
	{
			/* search backward for a white space */
		for ( i = maxc; i > 0 && !isspace (text[i]); i-- )
			;
		i++;	/* split *after* the space char, otherwise it */
			/* will be dropped by in the client */
		fprintf (out, "%.*s\r\n", i, text);

		text += i;	// move i chars forward in "text" ...
		len -= i;	// ... and reduce the length var

			// Now print the continuation char on the new line
		n = fprintf (out, "\t");	// even spaces are possible
		maxc = 75 - n;			// remaining line len
	}
	fprintf (out, "%s", text);	/* print the rest of the text */

	return n + len;
}

/*****************************************************************
**	print_description ()
**	
**	This prints a calendar text field. If the line would be longer
**	than 75, the text will be folded on space chars
**
**	DESCRIPTION:word ==> translation 
**		of the word
**	
*****************************************************************/
static	int	print_description (FILE *out, vc_t *v)
{
	int	n = 0;

	assert (v != NULL );

	n = fprintf (out, "DESCRIPTION:");
	n = printfolded (out, n, v->word);

	n = printfolded (out, n, ARROW_STRING);
	n = printfolded (out, n, v->trans);
	fprintf (out, "\r\n");

	return n;
}

/*****************************************************************
**
**	print_cal ()
**
**	prints whole calendar event
**
*****************************************************************/
int	print_cal (FILE *out, vc_t *v1, vc_t *v2, int nr)
{
	int	n = 0;
	struct	tm	date;
	const	char	*entrytype = "VEVENT";

	day (&date, firstday + DAYSEC (nr));

		/* don't print anything if v1 is not set and all on one entry */
	if ( qa_entry == 1 && v1 == NULL )
		return 0;

		/* VJOURNAL would be the right calendar type but has no client support */
	if ( vjournal )
		entrytype = "VJOURNAL";

	n += fprintf (out, "BEGIN:%s\r\n", entrytype);
	n += fprintf (out, "UID:vocab-%s-%02d@hznet.de\r\n", isodatestr (&date, 0), nr);

	n += fprintf (out, "DTSTAMP:%sT080000\r\n", isodatestr (&date, 0));
	n += fprintf (out, "DTSTART;VALUE=DATE:%s\r\n", isodatestr (&date, 0));
	n += fprintf (out, "SUMMARY:");
	n += printfolded (out, 8, v1 ? v1->word: "");
	n += fprintf (out, "\r\n");

	if ( qa_entry == 1 )	/* is answer on the same calender entry as the query? */
		n += print_description (out, v1);
	else if ( v2 )
		n += print_description (out, v2);

	n += fprintf (out, "CLASS:PUBLIC\r\n");
	n += fprintf (out, "TRANSP:TRANSPARENT\r\n");
	n += fprintf (out, "END:%s\r\n", entrytype);

	return n;
}

/*****************************************************************
**	print_cal_text ()
*****************************************************************/
int	print_cal_text (FILE *out, vc_t *v1, vc_t *v2, int nr)
{
	int	n = 0;
	struct	tm	date;

	day (&date, firstday + DAYSEC (nr));

	if ( qa_entry == 1 && v1 == NULL )
		return 0;

	if ( verbose )
		n += fprintf (out, "%2d: ", nr);

	n += fprintf (out, "%s: ", datestr (&date, 0));
	n += fprintf (out, "%20s", v1 ? v1->word: "");
	n += fprintf (out, "\t");
	if ( qa_entry == 1 )	/* is answer on the same calender entry as the query? */
	{
		n += fprintf (out, "%s", v1->word);
		n += fprintf (out, " ==> ");
		n += fprintf (out, "%s", v1->trans);
		n += print_description (out, v1);
	}
	else if ( v2 )
	{
		n += fprintf (out, "%s", v2->word);
		n += fprintf (out, " ==> ");
		n += fprintf (out, "%s", v2->trans);
	}
	n += fprintf (out, "\n");

	return n;
}

/*****************************************************************
**	main()
*****************************************************************/

#ifdef TEST
int	verbose;
int	qa_entry;
int	vjournal;

int	main (int argc, char *argv[])
{
	time_t		sec;
	struct	tm	date;
	int		i;
	
	if ( --argc )
		sec = startdate (*++argv);
	else
		sec = 0L;

	for ( i = 0; i < 14; i++ )
	{
		sec = day (&date, sec);
		//fprintf (stdout, "wday = %d\n", wday);
		fprintf (stdout, "%s", datestr (&date, 1));

		//day (&date, sec + DAYSEC( (7 - wday)+1 ));
		lastmonday (&date, sec);
		fprintf (stdout, "\t==> ");
		fprintf (stdout, "%s", datestr (&date, 1));
		fprintf (stdout, "\n");
		sec += DAYSEC(1);
	}
}
#endif
