/*****************************************************************
**
**	(@)#	vocab.c
**
**	Copyright (c) Feb 2017, Holger Zuleger HZNET. All rights reserved.
**
**	This software is open source
**
*****************************************************************/
# define	HAVE_GETOPT_LONG	1

# include <stdio.h>
# include <time.h>
# include <stdlib.h>
# include <string.h>
# include <ctype.h>
# include <assert.h>
#if defined(HAVE_GETOPT_LONG) && HAVE_GETOPT_LONG
# include <getopt.h>
#endif
# include "config.h"

#define extern
# include "vocab.h"
#undef extern
# include "vcl.h"
# include "regcards.h"
# include "cal.h"


static	int	pr_sequential (FILE *out, vcl_t *vcl, int (*outfct)());
static	int	pr_random (FILE *out, vcl_t *vcl, int n, int (*outfct)());

static	void	usage (const char *mesg);
static	int	print_text (FILE *out, vc_t *v1, vc_t *v2, int nr);

extern	int	optopt;
extern	int	opterr;
extern	int	optind;
extern	char	*optarg;

#  define	short_options	"hijac::qrsS:Vvn:u:"
#if defined(HAVE_GETOPT_LONG) && HAVE_GETOPT_LONG
static	struct	option	long_options[] = {
	{ "help",		no_argument,		NULL,	'h' },
	{ "version",		no_argument,		NULL,	'V' },
	{ "verbose",		no_argument,		NULL,	'v' },
	{ "invert",		no_argument,		NULL,	'i' },
	{ "journal",		no_argument,		NULL,	'j' },
	{ "regcard",		no_argument,		NULL,	'r' },
	{ "cal",		optional_argument,	NULL,	'c' },
	{ "sequential",		no_argument,		NULL,	's' },
	{ "qa-same",		no_argument,		NULL,	'q' },
	{ "size",		required_argument,	NULL,	'S' },
	{ "all",		no_argument,		NULL,	'a' },
	{ "number",		required_argument,	NULL,	'n' },
	{ "url",		required_argument,	NULL,	'u' },
	{  NULL,		no_argument,		NULL,	'\0' },
};
#endif

/* option vars */
const	char	*version = VERSION;
char	*progname = NULL;
int	nr = -1;		/* print all */
int	nrow = 10;		/* default grid	(row */
int	ncol = 2;		/*		(col) */
int	regcard = 0;		/* register card mode */
int	cal = 0;		/* calendar mode */
int	sequential = 0;		/* no random pickup */
int	qa_entry = 1;		/* query and answer on one the same calendar entry ? */
int	(*outf)() = print_text;		/* default output function */


static	int	print_text (FILE *out, vc_t *v1, vc_t *v2, int nr)
{
	int	n = 0;

	if ( v1 )
	{
		if ( verbose )
			n += fprintf (out, "%2d: ", nr);
		n += fprintf (out, "%s", v1->word);
		n += fprintf (out, " ==> ");
		n += fprintf (out, "%s", v1->trans);
		n += fprintf (out, "\n");
	}

	return n;
}

/*****************************************************************
**	pr_sequential()
**	call output function for every element in the vocabulary
*****************************************************************/
static	int	pr_sequential (FILE *out, vcl_t *vcl, int (*outfct)())
{
	int	i;
	int	n = 0;

	assert ( vcl != NULL );
	assert ( outfct != NULL );

	for ( i = 0; i < vcl->n; i++ )
		n += outfct (out, &vcl->list[i], NULL, i);

	return n;
}

/*****************************************************************
**	pr_random()
**	pick 'n' words randomly out of the vocab and call the outfunc
*****************************************************************/
static	int	pr_random (FILE *out, vcl_t *vcl, int n, int (*outfct)())
{
	int	elem[MAXVC];
	int	nelem;
	int	rnd;
	int	lastelem;
	int	bytes = 0;
	int	i;

	assert ( vcl != NULL );
	assert ( n <= vcl->n );
	assert ( outfct != NULL );

	for ( nelem = 0; nelem < vcl->n; nelem++ )
		elem[nelem] = nelem;

	lastelem = -1;
	for ( i = 0; n-- > 0; i++ )
	{
		rnd = random () % nelem;
		bytes += outfct (out, &vcl->list[elem[rnd]], lastelem < 0 ? NULL: &vcl->list[lastelem], i);

		lastelem = elem[rnd];
		elem[rnd] = elem[--nelem];
	}
	bytes += outfct (out, NULL, lastelem < 0 ? NULL: &vcl->list[lastelem], i);

	return bytes;
}

int	main(int argc, char *argv[])
{
	int	i;
	int	flag;
	int	opt_index;
	int	nr = -1;
	vcl_t	vcl;
	FILE	*fp;
	FILE	*out;
	char	*p;

	if ( (progname = strrchr (*argv, '/')) == NULL )
		progname = *argv;
	else
		progname++;

	vjournal = 0;		/* use VJOURNAL instead of VEVENT */
	qa_entry = 1;		/* query and answer on the same calendar entry ? */
	verbose = 0;
	invert = 0;
	ps = PS;
	vs = VS;
	url = NULL;

	opterr = 0;	/* prevent the printing of the getopt() build-in error messages */
#if defined(HAVE_GETOPT_LONG) && HAVE_GETOPT_LONG
	while ( (flag = getopt_long (argc, argv, short_options, long_options, &opt_index)) != -1 )
#else
	while ( (flag = getopt (argc, argv, short_options)) != -1 )
#endif
	{
		switch ( flag )
		{
		case 'V':	fprintf (stderr, "%s version %s\n", progname, version);
				fprintf (stderr, "compiled for a maximum of ");
				fprintf (stderr, "%d vocables\n", MAXVC);
				exit (0);
				break;
		case 'v':	verbose++;
				break;
		case 'a':	nr = MAXVC + 1;
				break;
		case 'i':	invert = 1;
				break;
		case 'j':	vjournal = 1;
				break;
		case 'n':	nr = atoi (optarg);
				break;
		case 'q':	qa_entry = !qa_entry;
				break;
		case 'r':	regcard = 1;
				outf = print_cardwords;
				break;
		case 'c':	cal = 1;
				nr = 13;
				if ( qa_entry )
					nr += 1;
				if ( optarg )
					setstartday (optarg);
				else
					setstartday (NULL);
				outf = print_cal;
				break;
		case 's':	sequential = 1;
				nr = MAXVC + 1;
				break;
		case 'S':	switch ( *optarg )
				{
				case 's':
					nrow = 20;
					ncol = 4;
					ps = 12;
					vs = 12;
					break;
				case 'm':
					nrow = 15;
					ncol = 3;
					ps = 14;
					vs = 15;
					break;
				case 'l':
					nrow = 10;
					ncol = 2;
					ps = 20;
					vs = 22;
					break;
				default:
					usage ("illegal size option\n");
				}
				break;
		case 'u':	url = optarg;
				break;
		case 'h':	usage (NULL);
				break;
		case ':':	fprintf (stderr, "option \"-%c\" requires an argument\n", optopt);
				usage ("");
				break;
		case '?':	if ( isprint (optopt) )
					fprintf (stderr, "%s: unknown option \"-%c\"\n", progname, optopt);
				else
					fprintf (stderr, "%s: unknown option char \\x%x\n", progname, optopt);
				usage ("");
				break;
		default:	abort();
		}
	}
	argc -= optind;
	argv += optind;

	if ( argc < 1 )
		usage ("Missing vocabulary file!");

		/* take the first filename as vocabulary name */
	snprintf (vocabname, sizeof (vocabname), "%s", argv[0]);
	while ( (p = strrchr (vocabname, '.')) != NULL )
		*p = '\0';

	vcl.n = 0;
	for ( i = 0; i < argc; i++ )
	{
		if ( (fp = fopen (argv[i], "r")) == NULL )
		{
			fprintf (stderr, "Can't open file \"%s\"\n", argv[i]);
			continue;
		}
		file2vcl (&vcl, invert, fp);
		fclose (fp);
	}

	out = stdout;
	if ( regcard )
	{
		if ( nr < 0 )
			nr = nrow * ncol;

		if ( !verbose )
			out = cardstream ();
	}

	if ( nr < 0 || nr > vcl.n )
		nr = vcl.n;

	if ( cal )
		cal_start (out);

	if ( sequential )
		pr_sequential (out, &vcl, outf);
	else
	{
		srandom ((unsigned int)time(NULL));
		pr_random (out, &vcl, nr, outf);
	}

	if ( cal )
		cal_end (out);

	if ( regcard )
	{
		print_cardheader (out);
		while ( nr > 0 )
		{
			print_cardpage (out, nrow, ncol, nr, 1);
			fprintf (out, ".bp\n");
			print_cardpage (out, nrow, ncol, nr, -1);
			nr -= nrow * ncol;
			if ( nr >  0 )
				fprintf (out, ".bp\n");
		}
	}

	if ( out != stdout )
		pclose (out);

	return 0;
}

void	error (const char *mesg)
{
	if ( mesg && *mesg )
		fprintf (stderr, "error: %s \n", mesg);
	exit (1);
}

static	void	usage (const char *mesg)
{
	if ( mesg && *mesg )
	{
		fprintf (stderr, "%s: error: ", progname);
		fprintf (stderr, "%s\n", mesg);
	}
	fprintf (stderr, "usage: %s -h|-V\n", progname);
	fprintf (stderr, "usage: %s [-s] [-a|-n #] [-i] file [...]\n", progname);
	fprintf (stderr, "usage: %s -r [-S small|medium|large] [-s] [-a|-n #] [-i] file [...]\n", progname);
	fprintf (stderr, "usage: %s -c[date] -u url [-q] [-s] [-a|-n #] [-i] file [...]\n", progname);

	fprintf (stderr, "\t-V, --version\t\t print out version and exit \n");
	fprintf (stderr, "\t-h, --help\t\t print out this help message \n");
	fprintf (stderr, "\t-v, --verbose\t\t give some hints about what's going on (don't start groff)\n");
	fprintf (stderr, "\t-i, --invert\t\t swap word and translation \n");
	fprintf (stderr, "\t-s, --sequential\t print all vocables in sequential order\n");
	fprintf (stderr, "\t-n, --number[=nr]\t print <nr> random vocables\n");
	fprintf (stderr, "\t-r, --regcard\t\t generate register card output for A4 page \n");
	fprintf (stderr, "\t-S, --size[=s|m|l]\t define # of cards per page (s=20*4, m=15*3, l=20*2)\n");
	fprintf (stderr, "\t-c, --cal\t\t generate (weekly) calendar output \n");
	fprintf (stderr, "\t-q, --qa-same\t\t query and answer on the same calendar sheet (default is yes) \n");
	fprintf (stderr, "\t-u, --url\t\t the calendar url \n");

	exit (0);
}
